/*
 * Based on Adobe's DDEOpen.cpp
 * Usage: acrocat file1 [file2] ...
 * Catenate all files into a single PDF file.
 *
 * e. woudenberg, fall 2001
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

// Constants/Declarations

const char *REG_INSTALL_KEY = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\Acrobat.exe";
const char *ACRO_DDESERVER = "acroview";
const char *ACRO_DDETOPIC = "control";

// globals

PROCESS_INFORMATION processInfo;
HCONV hConversation = NULL;
HSZ hszServerName, hszTopicName;
DWORD id = 0;
int load, store;
FILE *loadf, *storef;
int verbose, errors;
char *outfile = "out.pdf";
int waitsec = 2;

// forwards

HDDEDATA CALLBACK DDE_ProcessMessage (UINT uType, UINT uFmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
				      HDDEDATA hdata, DWORD dwData1, DWORD dwData2);
void LaunchAcrobat (PROCESS_INFORMATION *pProcessInfo);
int dde_init();
void dde_exec(char *ddeCmdBuf);
void dde_release();
int start_acroview();
void get_options(int *ac, char ***av);
void slave_loop();
void file_loop(int argc, char **argv);

void usage(char *name)
{
  printf("usage: %s [options] <file1> [file2] ...\n\
\n\
  Joins multiple files (tif, gif, pdf, etc. -- whatever can be opened by acrobat) \n\
  into a single pdf file\n\
\n\
  options: -o <file.pdf>  ; name Output file (default 'out.pdf')\n\
           -v             ; verbose, show DDE commands being sent to Acrobat\n\
           -s <file>      ; don't send DDE commands to Acrobat, just Store them to a file\n\
           -l <file>      ; don't generate Acrobat DDE commands, Load them from a file\n\
           -e             ; show DDE Error/Result values (not reliably reported, unfortunately)\n\
", name);
  exit(1);
}

int main(int argc, char* argv[])
{
  int e;

  get_options(&argc, &argv);

  if (!store) {
	LaunchAcrobat (&processInfo);
	if (!processInfo.hProcess)
	  return -1;
	
	Sleep(1000*waitsec);
	e = dde_init();
	if (e < 0) {
	  printf("Failed to initialize DDE. - DdeOpen - Error\n");
	  return -1;
	}
  }

  if (load)
	slave_loop();
  else
	file_loop(argc, argv);
  
  if (!store) {
	dde_release();
	TerminateProcess (processInfo.hProcess, 0);
  }

  return TRUE;
}

void dopen(char *s) {char cmd[1024]; sprintf(cmd, "[DocOpen(\"%s\")]", s); dde_exec(cmd);}
void dclose(char *s) {char cmd[1024]; sprintf(cmd, "[DocClose(\"%s\")]", s); dde_exec(cmd);}
void dsave(char *s) {char cmd[1024]; sprintf(cmd, "[DocSave(\"%s\")]", s); dde_exec(cmd);}
void dsaveas(char *s, char *t) {char cmd[1024]; sprintf(cmd, "[DocSaveAs(\"%s\",\"%s\")]", s, t); dde_exec(cmd);}
void dinsert(char *s, char *t) {char cmd[1024]; sprintf(cmd, "[DocInsertPages(\"%s\",PDLastPage,\"%s\")]", s, t); dde_exec(cmd);}
char  *ap(char *abs, char *p) {char *j; GetFullPathName(p, 1024, abs, &j); return abs;}

// extra "saveas" step due to broken(?) Adobe SDK: apparently cannot insert non-pdf files

void file_loop(int argc, char **argv)
{
  char bp[1024], cp[1024], dp[1024];

  ap(cp, argv[1]); dopen(cp);
  ap(bp, outfile); dsaveas(cp, bp);
  dclose(cp);
  dopen(bp);
  for (int ii = 2; ii < argc; ii++) {
	char *i = argv[ii];
	ap(cp, i); dopen(cp);
	ap(dp, i); strcat(dp, ".pdf"); dsaveas(cp, dp);
	dclose(cp);
	dinsert(bp, dp);
	dclose(dp); remove(dp);
  }
  Sleep(1000);
  dsave(bp);
  dclose(bp);
}

void slave_loop()
{
  while (1) {
	char ddeCmdBuf[1024];
	if (fgets(ddeCmdBuf, sizeof(ddeCmdBuf), loadf) == 0) break;
	if (ddeCmdBuf[0] == 0 || ddeCmdBuf[0] == '#') continue;
	dde_exec(ddeCmdBuf);
  }
}

void get_options(int *ac, char ***av)
{
  int argc = *ac;
  char **argv = *av;
  
  if (argc == 1)
	usage(argv[0]);

  while (argc > 1 && argv[1][0] == '-') {
	if (strcmp(argv[1], "-l") == 0) {
	  load++; loadf = stdin;
	  if (strcmp(argv[2], "-") != 0) {
		loadf = fopen(argv[2], "r");
		if (loadf == 0) {
		  fprintf(stderr, "couldn't open: %s\n", argv[2]);
		  exit(1);
		}
	  }
	  argc-=2; argv+=2;

	} else if (strcmp(argv[1], "-s") == 0) {
	  store++; storef = stdout;
	  if (strcmp(argv[2], "-") != 0) {
		storef = fopen(argv[2], "w");
		if (storef == 0) {
		  fprintf(stderr, "couldn't create: %s\n", argv[2]);
		  exit(1);
		}
	  }

	  argc-=2; argv+=2;

	} else if (strcmp(argv[1], "-w") == 0) {
	  waitsec = atoi(argv[2]);
	  argc-=2; argv+=2;

	} else if (strcmp(argv[1], "-o") == 0) {
	  outfile = argv[2];
	  argc-=2; argv+=2;

	} else if (strcmp(argv[1], "-v") == 0) {
	  verbose++;
	  argc--; argv++;

	} else if (strcmp(argv[1], "-e") == 0) {
	  errors++;
	  argc--; argv++;

	} else {
	  printf("Unknown switch %s\n", argv[1]);
	  usage(argv[0]);
	}
  }
  *ac = argc;
  *av = argv;
}


void dde_exec(char *ddeCmdBuf)
{
  DWORD dwResult;

  if (verbose)
	printf("%s\n", ddeCmdBuf);

  if (store)
	fprintf(storef, "%s\n", ddeCmdBuf);

  else {
	int e;
	// Execute the DDE Command.
	e = (int)DdeClientTransaction ((unsigned char *)ddeCmdBuf,
								   (DWORD)strlen(ddeCmdBuf), (HCONV)hConversation,
								   NULL, (UINT)CF_TEXT, (UINT)XTYP_EXECUTE, (DWORD)1000,
								   &dwResult);
	if (errors)
	  printf("%s = %d e=%d\n", ddeCmdBuf, (int)dwResult, e);
  }
}
	 
int dde_init()
{
  UINT retVal;
  int e = 0;

  retVal = DdeInitialize (&id, &DDE_ProcessMessage, APPCMD_CLIENTONLY, 0);
  if (DMLERR_NO_ERROR == retVal) {

    // Initialize DDE conversation with server.
    hszServerName = DdeCreateStringHandle (id, ACRO_DDESERVER, 0);
    hszTopicName = DdeCreateStringHandle (id, ACRO_DDETOPIC, 0);
    hConversation = DdeConnect (id, hszServerName, hszTopicName, NULL);

    retVal = DdeGetLastError (id);
    if (DMLERR_NO_ERROR != retVal) {
      printf("Could not connect to server. -- DdeOpen - Error\n");
	  e = -1;
	}
  }
  return e;
}

void dde_release()
{
  DdeDisconnect (hConversation);

  // Release resources.
  DdeFreeStringHandle (id, hszServerName);
  DdeFreeStringHandle (id, hszTopicName);
  DdeUninitialize (id);
}

HDDEDATA CALLBACK DDE_ProcessMessage (UINT uType, UINT uFmt, HCONV hconv, HSZ hsz1, HSZ hsz2,
				      HDDEDATA hdata, DWORD dwData1, DWORD dwData2)
{
  return NULL;
}

/* LaunchAcrobat
** ------------------------------------------------------
**
** Launches the Acrobat application.
**
** Fills the PROCESS_INFORMATION structure.
*/
void LaunchAcrobat (PROCESS_INFORMATION *pProcessInfo)
{
	long lRetCode;
	HKEY hkey;
	DWORD buffer[MAX_PATH];
	DWORD size = MAX_PATH;

	memset (pProcessInfo, 0, sizeof(PROCESS_INFORMATION));

	// Determine if a PDF viewer is installed.
	lRetCode = RegOpenKeyEx(HKEY_LOCAL_MACHINE, REG_INSTALL_KEY, 0, KEY_READ, &hkey);
	if (ERROR_SUCCESS == lRetCode) {

		// Get the path to the Acrobat executable.
		lRetCode = RegQueryValueEx (hkey, "", 0, 0, (unsigned char *)buffer, &size);
		RegCloseKey (hkey);

		STARTUPINFO startupInfo;
		memset (&startupInfo, 0, sizeof(STARTUPINFO));
		startupInfo.cb = sizeof(STARTUPINFO);
		startupInfo.wShowWindow = SW_HIDE;
		startupInfo.dwFlags = STARTF_USESHOWWINDOW;

		// Launch Acrobat.
		if (!CreateProcess((const char *)buffer, " /n /s /o", NULL, NULL, false,
			NORMAL_PRIORITY_CLASS, NULL, NULL, &startupInfo, pProcessInfo)) {

			char *errorBuf = 0;

			// Grab description of error.
			if (FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
								NULL, GetLastError(), 0, errorBuf, 0, NULL)) {

				// Advertise the error then abort.
				MessageBox (NULL, errorBuf, "BuildIndex - Error",MB_OK);
				LocalFree (errorBuf);
			}
		}
	} else
		// Advertise the error then abort.
		MessageBox (NULL, "Acrobat is not installed. Aborting...", "BuildIndex - Error",MB_OK);
}

